/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "screen_connector_launcher_service/tizen_remote_surface_internal.h"

namespace screen_connector {

TizenRemoteSurface::TizenRemoteSurface(struct tizen_remote_surface* raw,
                                       Direction direction,
                                       int32_t x, int32_t y,
                                       std::string shared_widget_info,
                                       uint32_t serial)
  : Handle<struct tizen_remote_surface*>(raw, true),
    direction_(direction),
    x_(x), y_(y),
    shared_widget_info_(std::move(shared_widget_info)),
    serial_(serial) {
}

TizenRemoteSurface::~TizenRemoteSurface() {
  if (GetRaw() != nullptr) {
    tizen_remote_surface_set_user_data(GetRaw(), NULL);
    tizen_remote_surface_unredirect(GetRaw());
    tizen_remote_surface_destroy(GetRaw());
  }
}

struct tizen_remote_surface* TizenRemoteSurface::GetHandle() const {
    return GetRaw();
  }

void TizenRemoteSurface::AddListener(IEventListener* listener) {
  static const struct tizen_remote_surface_listener rs_Listener = {
    BufferUpdatedCB,
    BufferMissingCB,
    BufferChangedCB,
    InputEventFilterCB,
  };

  listener_ = listener;
  tizen_remote_surface_add_listener(GetRaw(), &rs_Listener, this);
}

TizenRemoteSurface::IEventListener* TizenRemoteSurface::GetListener() {
  return listener_;
}

Direction TizenRemoteSurface::GetDirection() const {
  return direction_;
}

int32_t TizenRemoteSurface::GetPositionX() const {
  return x_;
}

int32_t TizenRemoteSurface::GetPositionY() const {
  return y_;
}

const std::string& TizenRemoteSurface::GetSharedWidgetInfo() const {
  return shared_widget_info_;
}

uint32_t TizenRemoteSurface::GetSerial() const {
  return serial_;
}

void TizenRemoteSurface::Redirect() {
  tizen_remote_surface_redirect(GetRaw());
}

void TizenRemoteSurface::BufferUpdatedCB(void* data,
                                         struct tizen_remote_surface* trs,
                                         struct wl_buffer* buffer,
                                         uint32_t time) {
}

void TizenRemoteSurface::BufferMissingCB(void* data,
                                         struct tizen_remote_surface* trs) {
  TizenRemoteSurface* rs = static_cast<TizenRemoteSurface*>(data);
  IEventListener* l = rs->GetListener();
  l->OnRawBufferMissing(rs->GetSerial());
}

void TizenRemoteSurface::BufferChangedCB(void* data,
                                         struct tizen_remote_surface* trs,
                                         uint32_t type,
                                         struct wl_buffer* tbm,
                                         int32_t img_fd,
                                         uint32_t img_file_size,
                                         uint32_t time,
                                         struct wl_array* keys) {
  TizenRemoteSurface* rs = static_cast<TizenRemoteSurface*>(data);
  IEventListener* l = rs->GetListener();
  l->OnRawBufferChanged(type, std::make_shared<WlBuffer>(tbm, trs),
      img_fd, img_file_size, time,
      rs->GetDirection(),
      rs->GetPositionX(),
      rs->GetPositionY(),
      rs->GetSharedWidgetInfo(),
      rs->GetSerial());
}

void TizenRemoteSurface::InputEventFilterCB(void* data,
                                            struct tizen_remote_surface* trs,
                                            uint32_t event_filter) {
  TizenRemoteSurface* rs = static_cast<TizenRemoteSurface*>(data);
  IEventListener* l = rs->GetListener();
  l->OnInputFilterChanged(event_filter, rs->GetSerial());
}

}  // namespace screen_connector
