/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "screen_connector_common/handle.h"
#include "screen_connector_launcher_service/tzsh_launcher_service_internal.h"
#include "screen_connector_launcher_service/log_internal.h"

namespace screen_connector {

TzshLauncherService::TzshLauncherService(tzsh_launcher_service_h raw)
  : Handle<tzsh_launcher_service_h>(raw, true) {
}

TzshLauncherService::~TzshLauncherService() {
  tzsh_launcher_service_destroy(GetRaw());
}

tzsh_launcher_service_h TzshLauncherService::GetHandle() const {
  return GetRaw();
}

void TzshLauncherService::SetListener(IEventListener* listener) {
  listener_ = listener;

  int ret = tzsh_launcher_service_prepare_cb_set(GetRaw(), PrepareCB,
      listener_);
  if (ret != TZSH_ERROR_NONE)
    _E("Failed to set prepare callback. error(%d)", ret);

  ret = tzsh_launcher_service_stop_cb_set(GetRaw(), StopCB, listener_);
  if (ret != TZSH_ERROR_NONE)
    _E("Failed to set stop callback. error(%d)", ret);

  ret = tzsh_launcher_service_error_cb_set(GetRaw(), ErrorCB, listener_);
  if (ret != TZSH_ERROR_NONE)
    _E("Failed to set error callback. error(%d)", ret);

  ret = tzsh_launcher_service_cleanup_cb_set(GetRaw(), CleanupCB, listener_);
  if (ret != TZSH_ERROR_NONE)
    _E("Failed to set cleanup callback. error(%d)", ret);
}

int TzshLauncherService::Launch(const std::string& app_id,
                                const std::string& inst_id,
                                int pid,
                                uint32_t serial) {
  return tzsh_launcher_service_launch(GetRaw(), app_id.c_str(),
      inst_id.c_str(), pid, serial);
}

int TzshLauncherService::LaunchWithSharedWidget(const std::string& app_id,
                                                const std::string& inst_id,
                                                int pid,
                                                uint32_t serial) {
  return tzsh_launcher_service_launch_with_shared_widget(GetRaw(),
      app_id.c_str(), inst_id.c_str(), pid, serial);
}

int TzshLauncherService::Launching(uint32_t serial) {
  return tzsh_launcher_service_launching(GetRaw(), serial);
}

int TzshLauncherService::LaunchCancel(uint32_t serial) {
  return tzsh_launcher_service_launch_cancel(GetRaw(), serial);
}

int TzshLauncherService::LaunchDone(uint32_t serial) {
  return tzsh_launcher_service_launch_done(GetRaw(), serial);
}

void TzshLauncherService::PrepareCB(void* data,
    tzsh_launcher_service_h service,
    tzsh_launcher_service_target_type_e target_type,
    uint32_t target_res_id,
    const char* target_path,
    const char* target_group,
    tzsh_launcher_service_direction_e direction,
    int32_t x,
    int32_t y,
    const char* target_widget_info,
    uint32_t serial) {
  IEventListener* listener = static_cast<IEventListener*>(data);
  std::string file_path(target_path ? target_path : "");
  std::string file_group(target_group ? target_group : "");
  std::string shared_widget_info(target_widget_info ? target_widget_info : "");
  listener->OnPrepareEvent(static_cast<uint32_t>(target_type),
      target_res_id, file_path, file_group,
      static_cast<Direction>(direction), x, y,
      shared_widget_info, serial);
}

void TzshLauncherService::StopCB(void* data,
                                 tzsh_launcher_service_h service,
                                 uint32_t serial) {
  IEventListener* listener = static_cast<IEventListener*>(data);
  listener->OnStopEvent(serial);
}

void TzshLauncherService::ErrorCB(void* data,
                                  tzsh_launcher_service_h service,
                                  tzsh_launcher_service_error_e error,
                                  uint32_t serial) {
  IEventListener* listener = static_cast<IEventListener*>(data);
  listener->OnErrorEvent(static_cast<Error>(error), serial);
}

void TzshLauncherService::CleanupCB(void* data,
                                    tzsh_launcher_service_h service,
                                    uint32_t serial) {
  IEventListener* listener = static_cast<IEventListener*>(data);
  listener->OnCleanupEvent(serial);
}

}  // namespace screen_connector
