/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <dlog.h>
#include <glib.h>
#include <unistd.h>

#include "screen_connector_remote_surface/remote_surface_manager_internal.h"
#include "screen_connector_remote_surface/remote_surface_implementation.h"
#include "screen_connector_remote_surface/remote_surface.h"
#include "screen_connector_remote_surface/aul_handle_mock_internal.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "SC_REMOTE_SURFACE"

namespace screen_connector {

uint32_t RemoteSurface::Impl::delayed_resuming_time_ = 0;
std::map<int, uint32_t> RemoteSurface::Impl::buffer_event_table_ = {
  {FILTER_NONE, TIZEN_REMOTE_SURFACE_CHANGED_BUFFER_EVENT_FILTER_NONE},
  {FILTER_TBM, TIZEN_REMOTE_SURFACE_CHANGED_BUFFER_EVENT_FILTER_TBM},
  {FILTER_IMAGE_FILE, TIZEN_REMOTE_SURFACE_CHANGED_BUFFER_EVENT_FILTER_IMAGE_FILE},
  {FILTER_ALL,
      TIZEN_REMOTE_SURFACE_CHANGED_BUFFER_EVENT_FILTER_TBM |TIZEN_REMOTE_SURFACE_CHANGED_BUFFER_EVENT_FILTER_IMAGE_FILE}
};

std::map<int, uint32_t>  RemoteSurface::Impl::input_type_table_ = {
  {MOUSE_UP_DOWN, TIZEN_REMOTE_SURFACE_INPUT_EVENT_FILTER_MOUSE_UP_DOWN},
  {MOUSE_MOVE_X, TIZEN_REMOTE_SURFACE_INPUT_EVENT_FILTER_MOUSE_MOVE_X},
  {MOUSE_MOVE_Y, TIZEN_REMOTE_SURFACE_INPUT_EVENT_FILTER_MOUSE_MOVE_Y},
  {MOUSE_MOVE_X_Y, TIZEN_REMOTE_SURFACE_INPUT_EVENT_FILTER_MOUSE_MOVE_X_Y},
  {MOUSE_WHEEL, TIZEN_REMOTE_SURFACE_INPUT_EVENT_FILTER_MOUSE_WHEEL_ALL},
  {TOUCH_UP_DOWN, TIZEN_REMOTE_SURFACE_INPUT_EVENT_FILTER_TOUCH_UP_DOWN},
  {TOUCH_MOVE_X, TIZEN_REMOTE_SURFACE_INPUT_EVENT_FILTER_TOUCH_MOVE_X},
  {TOUCH_MOVE_Y, TIZEN_REMOTE_SURFACE_INPUT_EVENT_FILTER_TOUCH_MOVE_Y},
  {TOUCH_MOVE_X_Y, TIZEN_REMOTE_SURFACE_INPUT_EVENT_FILTER_TOUCH_MOVE_X_Y},
  {TOUCH_CANCEL, TIZEN_REMOTE_SURFACE_INPUT_EVENT_FILTER_TOUCH_CANCEL_ALL},
  {KEY, TIZEN_REMOTE_SURFACE_INPUT_EVENT_FILTER_KEY_ALL}
};

void RemoteSurface::InitManager() {
  (void)RemoteSurfaceManager::GetInst();
}

void RemoteSurface::FiniManager() {
  RemoteSurfaceManager::GetInst().Dispose();
}

void RemoteSurface::SetDelayedResumingTime(uint32_t ms) {
  RemoteSurface::Impl::delayed_resuming_time_ = ms;
}

RemoteSurface::Impl::Impl(RemoteSurface* parent,
                          const std::string& id,
                          RemoteSurface::Type type,
                          std::shared_ptr<WlSurface> wl_surface, bool mock)
    : inst_id_(id), pid_(0), rid_(0), wl_surface_(std::move(wl_surface)),
      aul_handle_(mock ?
          new test::AulHandleMock(this, (aul_screen_type_e)type) :
          new AulHandle(this, (aul_screen_type_e)type)),
      type_(type), parent_(parent), mock_(mock) {
}

RemoteSurface::Impl::Impl(RemoteSurface* parent, int rid,
                          const std::string& id,
                          RemoteSurface::Type type,
                          std::shared_ptr<WlSurface> wl_surface, bool mock)
    : inst_id_(id), pid_(0), rid_(rid), wl_surface_(std::move(wl_surface)),
      aul_handle_(mock ?
          new test::AulHandleMock(this, (aul_screen_type_e)type) :
          new AulHandle(this, (aul_screen_type_e)type)),
      type_(type), parent_(parent), mock_(mock) {
}

void RemoteSurface::Impl::OnAppAdded(const std::string &appId,
                                     const std::string& instId,
                                     const int pid,
                                     const unsigned int surfaceId) {
  if (inst_id_ != instId)
    return;

  app_id_ = appId;
  inst_id_ = instId;
  rid_ = surfaceId;
  pid_ = pid;
  parent_->OnBufferAdded(appId, instId, pid);
  Redirect();
}

void RemoteSurface::Impl::OnAppRemoved(const std::string& appId,
                                       const std::string& instId,
                                       const int pid,
                                       const unsigned int surfaceId) {
  if (inst_id_ != instId)
    return;

  LOGW("removed %s", instId.c_str());
  parent_->OnBufferRemoved(appId, instId, pid);
}

void RemoteSurface::Impl::OnAppUpdated(const std::string &appId,
                                       const std::string& instId,
                                       const int pid,
                                       const unsigned int surfaceId) {
  if (inst_id_ != instId)
    return;

  if (rid_ != surfaceId)
    parent_->OnBufferRemoved(appId, instId, pid);

  inst_id_ = instId;
  rid_ = surfaceId;
  pid_ = pid;
  parent_->OnBufferAdded(appId, instId, pid);
  Redirect();
}

RemoteSurface::RemoteSurface(const std::string& id, Type type,
      std::shared_ptr<WlSurface> wl_surface, bool mock)
    : impl_(new Impl(this, id, type, std::move(wl_surface), mock)) {
}

RemoteSurface::RemoteSurface(int rid, const std::string& id,
                    Type type, std::shared_ptr<WlSurface> wl_surface, bool mock)
    : impl_(new Impl(this, rid, id, type, std::move(wl_surface), mock)) {
  impl_->Redirect();
}

RemoteSurface::~RemoteSurface() = default;
RemoteSurface::RemoteSurface(RemoteSurface&&) noexcept = default;
RemoteSurface& RemoteSurface::operator=(RemoteSurface&&) noexcept = default;

void RemoteSurface::Impl::ClearResumingTimer() {
  if (resuming_timer_ > 0) {
    g_source_remove(resuming_timer_);
    resuming_timer_ = 0;
  }
}

void RemoteSurface::Impl::SetResumingTimer() {
  ClearResumingTimer();
  resuming_timer_ = g_timeout_add(delayed_resuming_time_,
      [](gpointer user_data)->gboolean {
    RemoteSurface::Impl* rs = static_cast<RemoteSurface::Impl*>(user_data);

    rs->surface_->TransferVisibility(true);
    rs->resuming_timer_ = 0;

    return G_SOURCE_REMOVE;
  }, this);
}

void RemoteSurface::SendVisibility(bool visible) {
  if (impl_->freeze_)
    return;

  if (RemoteSurface::Impl::delayed_resuming_time_ == 0 ||
      GetType() == WATCH) {
    if (!visible)
      SendTouchCancel();
    LOGI("normal resume %d, %s", visible, impl_->inst_id_.c_str());
    if (visible) {
      impl_->surface_->TransferVisibility(true);
    } else {
      impl_->surface_->TransferVisibility(false);
    }
    return;
  }

  if (visible) {
    impl_->SetResumingTimer();
  } else {
    impl_->ClearResumingTimer();
    SendTouchCancel();
    impl_->surface_->TransferVisibility(false);
  }
}

void RemoteSurface::SendTouchCancel() {
  impl_->surface_->TransferTouchCancel();
}

void RemoteSurface::Impl::OnRawBufferMissing() {
  parent_->OnBufferRemoved(parent_->GetAppId(), parent_->GetInstId(),
                           parent_->GetPid());
}

void RemoteSurface::Impl::OnRawBufferChanged(int type,
                                             std::shared_ptr<WlBuffer> tbm,
                                             int fd, uint32_t size,
                                             uint32_t time) {
  if (bind_surface_.get() != nullptr || block_)
    return;

  LOGI("resource_id : %s", parent_->GetInstId().c_str());
  parent_->OnBufferChanged(type, tbm, fd, size, time);
}

void RemoteSurface::Impl::OnInputFilterChanged(uint32_t event_filter) {
  input_event_filter_ = event_filter;
  LOGI("event filter updated:%d", event_filter);
}

void RemoteSurface::Impl::Redirect() {
  surface_.reset(RemoteSurfaceManager::GetInst(mock_)
    .CreateTRS(rid_, wl_surface_, mock_));

  LOGI("resource_id : %d, (%p), %s",
      rid_, surface_->GetHandle(), inst_id_.c_str());
  surface_->AddListener(this);
  surface_->Redirect();
  parent_->SetChangedEventFilter(changed_event_filter_);

  if (bind_surface_.get() != nullptr)
    RemoteSurfaceManager::GetInst().Bind(*surface_, *bind_surface_);
}

void RemoteSurface::Bind(std::shared_ptr<WlSurface> surface) {
  SendTouchCancel();
  RemoteSurfaceManager::GetInst().Bind(*(impl_->surface_), *surface);
  impl_->bind_surface_ = surface;
  OnBind();
}

void RemoteSurface::Unbind() {
  RemoteSurfaceManager::GetInst().Unbind(*(impl_->surface_));
  impl_->bind_surface_.reset();
}

bool RemoteSurface::IsBound() const {
  if (impl_->bind_surface_.get() != nullptr)
    return true;

  return false;
}

void RemoteSurface::Freeze(bool visible) {
  SendVisibility(visible);
  impl_->freeze_ = true;
}

void RemoteSurface::SendMouseUp(int x, int y, unsigned int timestamp) {
  impl_->surface_->TransferMouseUp(x, y, timestamp);
}

void RemoteSurface::SetChangedEventFilter(ChangedEventFilter filter) {
  impl_->surface_->SetChangedBufferEventFilter(impl_->buffer_event_table_[filter]);
  impl_->changed_event_filter_ = filter;
  LOGI("set filter [%d] for [%s]",
      impl_->changed_event_filter_, GetInstId().c_str());
}

void RemoteSurface::SetBlock(bool is_block) {
  impl_->block_ = is_block;
}

bool RemoteSurface::IsDisabledInput(InputType type) const {
  return ((impl_->input_event_filter_ & impl_->input_type_table_[type]) > 0);
}

void RemoteSurface::OnBufferAdded(const std::string& appId,
                                  const std::string& instId, int pid) {
}

void RemoteSurface::OnBufferRemoved(const std::string& appId,
                                    const std::string& instId, int pid) {
}

void RemoteSurface::OnBufferChanged(int type, std::shared_ptr<WlBuffer> tbm,
                                    int fd, uint32_t size, uint32_t time) {
}

void RemoteSurface::OnBind() {
}

const std::string& RemoteSurface::GetAppId() const {
  return impl_->app_id_;
}

const std::string& RemoteSurface::GetInstId() const {
  return impl_->inst_id_;
}

int RemoteSurface::GetPid() const {
  return impl_->pid_;
}

int RemoteSurface::GetRid() const {
  return impl_->rid_;
}

RemoteSurface::Type RemoteSurface::GetType() const {
  return impl_->type_;
}

bool RemoteSurface::IsFrozen() const {
  return impl_->freeze_;
}

ITRS* RemoteSurface::GetRaw() {
  return impl_->surface_.get();
}

void RemoteSurface::Thaw() {
  impl_->freeze_ = false;
}

RemoteSurface::ChangedEventFilter RemoteSurface::GetChangedEventFilter() const {
  return impl_->changed_event_filter_;
}

void RemoteSurface::ClearRemoteRender() {
  impl_->renderable_ = false;
  impl_->surface_->SetRemoteRender(false);
}

int RemoteSurface::SetRemoteRender() {
  LOGI("send update request to %s", impl_->inst_id_.c_str());
  impl_->surface_->SetRemoteRender(true);
  impl_->renderable_ = true;

  return aul_screen_connector_send_update_request(impl_->app_id_.c_str(),
    impl_->inst_id_.c_str());
}

}  // namespace screen_connector
